var urllib      = require('url');
var querystring = require('querystring');
var sax         = require('sax');
var request     = require('./request');
var util        = require('./util');
var extras      = require('./info-extras');
var sig         = require('./sig');
var cache       = require('./cache');


var VIDEO_URL = 'https://www.youtube.com/watch?v=';
var EMBED_URL = 'https://www.youtube.com/embed/';
var VIDEO_EURL = 'https://youtube.googleapis.com/v/';
var INFO_HOST = 'www.youtube.com';
var INFO_PATH = '/get_video_info';


/**
 * Gets info from a video without getting additional formats.
 *
 * @param {string} id
 * @param {Object} options
 * @param {Function(Error, Object)} callback
 */
function getBasicInfo(id, options, callback) {

  id = util.getURLVideoID(id);

  // Try getting config from the video page first.
  var params = 'hl=' + (options.lang || 'en');
  var url = VIDEO_URL + id + '&' + params +
    '&bpctr=' + Math.ceil(Date.now() / 1000);

  // Remove header from watch page request.
  // Otherwise, it'll use a different framework for rendering content.
  /*
  var reqOptions = Object.assign({}, options.requestOptions);
  reqOptions.headers = Object.assign({}, reqOptions.headers, {
    'User-Agent': ''
  });
  */
  var myrequest = options.request || request;

  myrequest(url, options.requestOptions, function(err, body) {
		if (err) return callback(err);

    // Check if there are any errors with this video page.
    var unavailableMsg = util.between(body, '<div id="player-unavailable"', '>');
    if (unavailableMsg &&
        !/\bhid\b/.test(util.between(unavailableMsg, 'class="', '"'))) {
      // Ignore error about age restriction.
      if (!body.includes('<div id="watch7-player-age-gate-content"')) {
        return callback(new Error(util.between(body,
          '<h1 id="unavailable-message" class="message">', '</h1>').trim()));
      }
    }



    // Parse out additional metadata from this page.
    var additional = {
      // Get the author/uploader.
      //author: extras.getAuthor(body),

      // Get the day the vid was published.
      //published: extras.getPublished(body),

      // Get description.
      //description: extras.getVideoDescription(body),

      // Get media info.
      //media: extras.getVideoMedia(body),

      // Get related videos.
      //related_videos: extras.getRelatedVideos(body),
    };

    var jsonStr = util.between(body, 'ytplayer.config = ', '</script>');
    var config;
    if (jsonStr) {
      //config = jsonStr.slice(0, jsonStr.lastIndexOf(';ytplayer.load')); 
      config = jsonStr.slice(0, jsonStr.indexOf(';ytplayer.')); 
	  //console.log(config);
      gotConfig(id, options, additional, config, false, callback);

    } else {
      // If the video page doesn't work, maybe because it has mature content.
      // and requires an account logged in to view, try the embed page.
      url = EMBED_URL + id + '?' + params;
      myrequest(url, options.requestOptions, function(err, body) {
        if (err) return callback(err);
        config = util.between(body, 't.setConfig({\'PLAYER_CONFIG\': ', /\}(,'|\}\);)/);
        gotConfig(id, options, additional, config, true, callback);
      });
    }
  });
};


/**
 * @param {Object} info
 * @return {Array.<Object>}
 */
var parseFormats = function(info) {
  var formats = [];
  if (info.player_response.streamingData) {
    if (info.player_response.streamingData.formats) {
      formats = formats.concat(info.player_response.streamingData.formats);
    }
    if (info.player_response.streamingData.adaptiveFormats) {
      formats = formats.concat(info.player_response.streamingData.adaptiveFormats);
    }
  }
  return formats;
};


/**
 * @param {Object} id
 * @param {Object} options
 * @param {Object} additional
 * @param {Object} config
 * @param {boolean} fromEmbed
 * @param {Function(Error, Object)} callback
 */
var gotConfig = function(id, options, additional, config, fromEmbed, callback) {
  if (!config) {
    return callback(new Error('Could not find player config'));
  }
  try {
    config = JSON.parse(config + (fromEmbed ? '}' : ''));
  } catch (err) {
    return callback(new Error('Error parsing config: ' + err.message));
  }
  
  var url = urllib.format({
    protocol: 'https',
    host: INFO_HOST,
    pathname: INFO_PATH,
    query: {
      video_id: id,
      eurl: VIDEO_EURL + id,
      ps: 'default',
      gl: 'US',
      hl: (options.lang || 'en'),
      sts: (config.sts?config.sts:""),
    },
  });
 
  var myrequest = options.request || request;
  myrequest(url, options.requestOptions, function(err, body) {
    if (err) return callback(err);

    var info = querystring.parse(body);
    var player_response = config.args.player_response || info.player_response;

    if (info.status === 'fail') {
      return callback(
        new Error('Code '+ info.errorcode+" : "+util.stripHTML(info.reason)));
    } else try {
      info.player_response = JSON.parse(player_response);
    } catch (err) {
      return callback(
        new Error('Error parsing `player_response`: ' + err.message));
    }

    var playability = info.player_response.playabilityStatus;
    if (playability && playability.status === 'UNPLAYABLE') {
      return callback(new Error(playability.reason));
    }

    info.formats = parseFormats(info);

/*
    // Add additional properties to info.
    Object.assign(info, additional, {
      video_id: id,

      // Give the standard link to the video.
      video_url: VIDEO_URL + id,

      // Copy over a few props from `player_response.videoDetails`
      // for backwards compatibility.
      title: info.player_response.videoDetails && info.player_response.videoDetails.title,
      length_seconds: info.player_response.videoDetails && info.player_response.videoDetails.lengthSeconds,
    });
*/
	info.video_id = id;
	info.video_url = VIDEO_URL + id;
	info.title = info.player_response.videoDetails && info.player_response.videoDetails.title;
	info.length_seconds = info.player_response.videoDetails && info.player_response.videoDetails.lengthSeconds;

    info.age_restricted = fromEmbed;
    info.html5player = config.assets.js;

    callback(null, info);
  });
};


/**
 * Gets info from a video additional formats and deciphered URLs.
 *
 * @param {string} id
 * @param {Object} options
 * @param {Function(Error, Object)} callback
 */
module.exports = function getFullInfo(id, options, callback) {
  if (typeof options === 'function') {
    callback = options;
    options = {};
  } else if (!options) {
    options = {};
  }
  return getBasicInfo(id, options, function(err, info) {

    if (err) return callback(err);
    var hasManifest =
      info.player_response && info.player_response.streamingData && (
        info.player_response.streamingData.dashManifestUrl ||
        info.player_response.streamingData.hlsManifestUrl
      );
    if (info.formats.length || hasManifest) {
      var html5playerfile = urllib.resolve(VIDEO_URL, info.html5player);
      sig.getTokens(html5playerfile, options, function(err, tokens) {
        if (err) return callback(err);

        sig.decipherFormats(info.formats, tokens, options.debug);

		info.formats.sort(util.sortFormats);
  	    info.full = true;
	    callback(null, info);

		/*		
        var funcs = [];

        if (hasManifest && info.player_response.streamingData.hlsManifestUrl) {
          var url = info.player_response.streamingData.hlsManifestUrl;
          funcs.push(getM3U8.bind(null, url, options));
        }


        if (hasManifest && info.player_response.streamingData.dashManifestUrl) {
          var url = info.player_response.streamingData.dashManifestUrl;
          funcs.push(getDashManifest.bind(null, url, options));
        }


        util.parallel(funcs, function(err, results) {
          if (err) return callback(err);
          if (results[0]) { mergeFormats(info, results[0]); }
          if (results[1]) { mergeFormats(info, results[1]); }
          //info.formats = info.formats.map(util.addFormatMeta);
          info.formats.sort(util.sortFormats);
          info.full = true;
          callback(null, info);
        });
		*/		
      });
    } else {
      callback(new Error('This video is unavailable'));
    }
  });
};


/**
 * Merges formats from DASH or M3U8 with formats from video info page.
 *
 * @param {Object} info
 * @param {Object} formatsMap
 */
var mergeFormats = function(info, formatsMap) {
  info.formats.forEach(function(f) {
    formatsMap[f.itag] = f;
  });
  info.formats = Object.keys(formatsMap).map(function(e) { return formatsMap[e]});
  //info.formats = Object.values(formatsMap);
};


/**
 * Gets additional DASH formats.
 *
 * @param {string} url
 * @param {Object} options
 * @param {Function(!Error, Array.<Object>)} callback
 */
var getDashManifest = function(url, options, callback) {
  var formats = {};
  var currentFormat = null;

  var parser = sax.parser(false);
  parser.onerror = callback;
  parser.onopentag = function(node) {
    if (node.name === 'REPRESENTATION') {
      //var itag = node.attributes.ID;
      //formats[itag] = { itag, url };
      var itag = node.attributes.ID;
      var meta = FORMATS[itag];
      currentFormat = { itag: itag, url: url };
      for (var key in meta) {
        currentFormat[key] = meta[key];
      }
      formats[itag] = currentFormat;
    }
  };
  parser.onend = function() { callback(null, formats); };

  var req = request(urllib.resolve(VIDEO_URL, url), options.requestOptions);
  //req.setEncoding('utf8');
  req.on('error', callback);
  req.on('data', function(chunk) { parser.write(chunk); });
  req.on('end', parser.close.bind(parser));
};


/**
 * Gets additional formats.
 *
 * @param {string} url
 * @param {Object} options
 * @param {Function(!Error, Array.<Object>)} callback
 */
var getM3U8 = function(url, options, callback) {
  url = urllib.resolve(VIDEO_URL, url);
  request(url, options.requestOptions, function(err, body) {
    if (err) return callback(err);

    var formats = {};
    body
      .split('\n')
      .filter(function(line) { /https?:\/\//.test(line)})
      .forEach(function(line) {
        var itag = line.match(/\/itag\/(\d+)\//)[1];
        formats[itag] = { itag: itag, url: line };
      });
    callback(null, formats);
  });
};


// Cached for getting basic/full info.
//cache = new Cache();


// Cache get info functions.
// In case a user wants to get a video's info before downloading.


// Export a few helpers.
exports.validateID = util.validateID;
exports.validateURL = util.validateURL;
exports.getURLVideoID = util.getURLVideoID;
exports.getVideoID = util.getVideoID;
